@@ -12,13 +12,13 @@
 /*	void	cleanup_out(state, type, data, len)
 /*	CLEANUP_STATE *state;
 /*	int	type;
-/*	char	*data;
-/*	int	len;
+/*	const char *data;
+/*	ssize_t	len;
 /*
 /*	void	cleanup_out_string(state, type, str)
 /*	CLEANUP_STATE *state;
 /*	int	type;
-/*	char	*str;
+/*	const char *str;
 /*
 /*	void	CLEANUP_OUT_BUF(state, type, buf)
 /*	CLEANUP_STATE *state;
@@ -28,7 +28,11 @@
 /*	void	cleanup_out_format(state, type, format, ...)
 /*	CLEANUP_STATE *state;
 /*	int	type;
-/*	char	*format;
+/*	const char *format;
+/*
+/*	void	cleanup_out_header(state, buf)
+/*	CLEANUP_STATE *state;
+/*	VSTRING	*buf;
 /* DESCRIPTION
 /*	This module writes records to the output stream.
 /*
@@ -47,6 +51,10 @@
 /*
 /*	cleanup_out_format() formats its arguments and writes
 /*	the result as a record.
+/*
+/*	cleanup_out_header() outputs a multi-line header as records
+/*	of the specified type. The input is expected to be newline
+/*	separated (not newline terminated), and is modified.
 /* LICENSE
 /* .ad
 /* .fi
@@ -71,6 +79,7 @@
 #include <msg.h>
 #include <vstring.h>
 #include <vstream.h>
+#include <split_at.h>
 
 /* Global library. */
 
@@ -78,6 +87,7 @@
 #include <rec_type.h>
 #include <cleanup_user.h>
 #include <mail_params.h>
+#include <lex_822.h>
 
 /* Application-specific. */
 
@@ -85,7 +95,7 @@
 
 /* cleanup_out - output one single record */
 
-void    cleanup_out(CLEANUP_STATE *state, int type, char *string, int len)
+void    cleanup_out(CLEANUP_STATE *state, int type, const char *string, ssize_t len)
 {
     int     err = 0;
 
@@ -104,6 +114,8 @@ void    cleanup_out(CLEANUP_STATE *state, int type, char *string, int len)
 
 #define TEXT_RECORD(t)	((t) == REC_TYPE_NORM || (t) == REC_TYPE_CONT)
 
+    if (var_line_limit <= 0)
+	msg_panic("cleanup_out: bad line length limit: %d", var_line_limit);
     do {
 	if (len > var_line_limit && TEXT_RECORD(type)) {
 	    err = rec_put(state->dst, REC_TYPE_CONT, string, var_line_limit);
@@ -129,14 +141,14 @@ void    cleanup_out(CLEANUP_STATE *state, int type, char *string, int len)
 
 /* cleanup_out_string - output string to one single record */
 
-void    cleanup_out_string(CLEANUP_STATE *state, int type, char *string)
+void    cleanup_out_string(CLEANUP_STATE *state, int type, const char *string)
 {
     cleanup_out(state, type, string, strlen(string));
 }
 
 /* cleanup_out_format - output one formatted record */
 
-void    cleanup_out_format(CLEANUP_STATE *state, int type, char *fmt,...)
+void    cleanup_out_format(CLEANUP_STATE *state, int type, const char *fmt,...)
 {
     static VSTRING *vp;
     va_list ap;
@@ -148,3 +160,41 @@ void    cleanup_out_format(CLEANUP_STATE *state, int type, char *fmt,...)
     va_end(ap);
     CLEANUP_OUT_BUF(state, type, vp);
 }
+
+/* cleanup_out_header - output one multi-line header as a bunch of records */
+
+void    cleanup_out_header(CLEANUP_STATE *state, VSTRING *header_buf)
+{
+    char   *start = vstring_str(header_buf);
+    char   *line;
+    char   *next_line;
+
+    /*
+     * Prepend a tab to continued header lines that went through the address
+     * rewriting machinery. See cleanup_fold_header(state) below for the form
+     * of such header lines. NB: This code destroys the header. We could try
+     * to avoid clobbering it, but we're not going to use the data any
+     * further.
+     * 
+     * XXX We prefer to truncate a header at the last line boundary before the
+     * header size limit. If this would undershoot the limit by more than
+     * 10%, we truncate between line boundaries to avoid losing too much
+     * text. This "unkind cut" may result in syntax errors and may trigger
+     * warnings from down-stream MTAs.
+     */
+    for (line = start; line; line = next_line) {
+	next_line = split_at(line, '\n');
+	if ((next_line ? next_line - 1 : line + strlen(line))
+	    > start + var_header_limit) {
+	    if (line - start > 0.9 * var_header_limit)	/* nice cut */
+		break;
+	    start[var_header_limit] = 0;	/* unkind cut */
+	    next_line = 0;
+	}
+	if (line == start || IS_SPACE_TAB(*line)) {
+	    cleanup_out_string(state, REC_TYPE_NORM, line);
+	} else {
+	    cleanup_out_format(state, REC_TYPE_NORM, "\t%s", line);
+	}
+    }
+}
